@@ -5,29 +5,24 @@ The application classes in this module are installed as executables via
 setuptools entry points.
 """
 
-from __future__ import absolute_import
-from __future__ import unicode_literals
-from __future__ import print_function
-
 # import udiskie.depend first - for side effects!
 from .depend import has_Notify, has_Gtk, _in_X
 
-import sys
 import inspect
 import logging.config
 import traceback
 
-from docopt import docopt, DocoptExit
-
 from gi.repository import GLib
 
+from docopt import docopt, DocoptExit
+
 import udiskie
 import udiskie.config
 import udiskie.mount
-import udiskie.compat
-from .async_ import AsyncList, Coroutine, Return, Async
-from .common import extend, str2unicode, ObjDictView
+import udiskie.udisks2
+from .common import extend, ObjDictView
 from .locale import _
+from .async_ import Future, ensure_future, gather
 
 
 __all__ = [
@@ -37,51 +32,7 @@ __all__ = [
 ]
 
 
-def deprecation_warning(text):
-    """Show a deprecation warning."""
-    # NOTE: not using `warnings.warn(text, DeprecationWarning)`, because that
-    # requires starting the python interpreter to be started with `-Wd`, which
-    # in turn shows more warnings about removal of Gtk.StatusIcon.
-    log = logging.getLogger("udiskie.DeprecationWarning")
-    log.warning(_("Deprecation warning: {}", text))
-
-
-@Coroutine.from_generator_function
-def get_backend(version=None):
-    """
-    Return UDisks service.
-
-    :param int version: requested UDisks backend version
-    :returns: UDisks service wrapper object
-    :raises GLib.GError: if unable to connect to UDisks dbus service.
-    :raises ValueError: if the version is invalid
-
-    If ``version`` has a false truth value, try to connect to UDisks2 and
-    fall back to UDisks1 if not available.
-    """
-    if not version:
-        try:
-            daemon = yield get_backend(2)
-        except GLib.GError:
-            log = logging.getLogger(__name__)
-            log.warning(_('Failed to connect UDisks2 dbus service..\n'
-                          'Falling back to UDisks1.'))
-            daemon = yield get_backend(1)
-    elif version == 1:
-        import udiskie.udisks1
-        daemon = yield udiskie.udisks1.Daemon.create()
-        deprecation_warning(_(
-            'Using UDisks1. Support will be discontinued '
-            'in the next major version of udiskie.'))
-    elif version == 2:
-        import udiskie.udisks2
-        daemon = yield udiskie.udisks2.Daemon.create()
-    else:
-        raise ValueError(_("UDisks version not supported: {0}!", version))
-    yield Return(daemon)
-
-
-class Choice(object):
+class Choice:
 
     """Mapping of command line arguments to option values."""
 
@@ -109,7 +60,7 @@ def Switch(name):
                    '--no-' + name: False})
 
 
-class Value(object):
+class Value:
 
     """Option which is given as value of a command line argument."""
 
@@ -119,10 +70,10 @@ class Value(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return str2unicode(args[self._name])
+        return args[self._name]
 
 
-class OptionalValue(object):
+class OptionalValue:
 
     def __init__(self, name):
         """Set argument name."""
@@ -131,40 +82,38 @@ class OptionalValue(object):
 
     def __call__(self, args):
         """Get the value of the command line argument."""
-        return self._choice(args) and str2unicode(args[self._name])
+        return self._choice(args) and args[self._name]
 
 
 class SelectLevel(logging.Filter):
+
     def __init__(self, level):
         self.level = level
+
     def filter(self, record):
         return record.levelno == self.level
 
 
-class _EntryPoint(object):
+class _EntryPoint:
 
     """
     Abstract base class for program entry points.
 
-    Concrete implementations need to implement :meth:`run` and extend
-    :meth:`finalize_options` to be usable with :meth:`main`. Furthermore
-    the docstring of any concrete implementation must be usable with
-    docopt. :ivar:`name` must be set to the name of the CLI utility.
+    Implementations need to
+
+    - implement :meth:`_init`
+    - provide a docstring
+    - extend :cvar:`option_defaults` and :cvar:`option_rules`.
     """
 
     option_defaults = {
         'log_level': logging.INFO,
-        'udisks_version': None,
     }
 
     option_rules = {
         'log_level': Choice({
             '--verbose': logging.DEBUG,
             '--quiet': logging.ERROR}),
-        'udisks_version': Choice({
-            '--udisks-auto': 0,
-            '--use-udisks1': 1,
-            '--use-udisks2': 2}),
     }
 
     usage_remarks = _("""
@@ -175,14 +124,9 @@ class _EntryPoint(object):
     """)
 
     def __init__(self, argv=None):
-        """
-        Parse command line options, read config and initialize members.
-
-        :param list argv: command line parameters
-        """
-        udiskie.compat.patch_print_unicode()
+        """Parse command line options, read config and initialize members."""
         # parse program options (retrieve log level and config file name):
-        args = docopt(self.usage, version=self.name + ' ' + self.version)
+        args = docopt(self.usage, version='udiskie ' + self.version)
         default_opts = self.option_defaults
         program_opts = self.program_options(args)
         # initialize logging configuration:
@@ -193,7 +137,8 @@ class _EntryPoint(object):
             'disable_existing_loggers': False,
             'formatters': {
                 'plain':  {'format': _('%(message)s')},
-                'detail': {'format': _('%(levelname)s [%(asctime)s] %(name)s: %(message)s')},
+                'detail': {'format': _(
+                    '%(levelname)s [%(asctime)s] %(name)s: %(message)s')},
             },
             'filters': {
                 'info': {'()': 'udiskie.cli.SelectLevel', 'level': logging.INFO},
@@ -227,21 +172,9 @@ class _EntryPoint(object):
         # initialize instance variables
         self.config = config
         self.options = options
-        self.exit_status = 0
-        if sys.version_info < (3,5):
-            deprecation_warning(_(
-                "Running on python {}.{}. The next major version of udiskie "
-                "will require at least python 3.5!",
-                *sys.version_info[:2]))
 
     def program_options(self, args):
-        """
-        Fully initialize Daemon object.
-
-        :param dict args: arguments as parsed by docopt
-        :returns: options from command line
-        :rtype: dict
-        """
+        """Get program options from docopt parsed options."""
         options = {}
         for name, rule in self.option_rules.items():
             val = rule(args)
@@ -251,70 +184,52 @@ class _EntryPoint(object):
 
     @classmethod
     def main(cls, argv=None):
-        """
-        Run program.
-
-        :param list argv: command line parameters
-        :returns: program exit code
-        :rtype: int
-        """
+        """Run program. Returns program exit code."""
         return cls(argv).run()
 
     @property
     def version(self):
-        """Get the version from setuptools metadata."""
+        """Version from setuptools metadata."""
         return udiskie.__version__
 
     @property
     def usage(self):
-        """Get the full usage string."""
+        """Full usage string."""
         return inspect.cleandoc(self.__doc__ + self.usage_remarks)
 
-    @property
-    def name(self):
-        """Get the name of the CLI utility."""
-        raise NotImplementedError()
-
     def _init(self):
-        """
-        Fully initialize Daemon object.
-
-        :returns: the main application task
-        :rtype: Async
-        """
+        """Return the application main task as Future."""
         raise NotImplementedError()
 
     def run(self):
-        """
-        Run the main loop.
-
-        :returns: exit code
-        :rtype: int
-        """
+        """Run the main loop. Returns exit code."""
+        self.exit_code = 1
         self.mainloop = GLib.MainLoop()
-        self._start_async_tasks()
         try:
+            future = ensure_future(self._start_async_tasks())
+            future.callbacks.append(self.set_exit_code)
             self.mainloop.run()
-            return self.exit_status
+            return self.exit_code
         except KeyboardInterrupt:
             return 1
 
-    @Coroutine.from_generator_function
-    def _start_async_tasks(self):
+    def set_exit_code(self, exit_code):
+        self.exit_code = exit_code
+
+    async def _start_async_tasks(self):
         """Start asynchronous operations."""
         try:
-            self.udisks = yield get_backend(self.options['udisks_version'])
-            results = yield self._init()
-            if not all(results):
-                self.exit_status = 1
+            self.udisks = await udiskie.udisks2.Daemon.create()
+            results = await self._init()
+            return 0 if all(results) else 1
         except Exception:
-            self.exit_status = 1
-            # Print the stack trace only up to the current level:
             traceback.print_exc()
-        self.mainloop.quit()
+            return 1
+        finally:
+            self.mainloop.quit()
 
 
-class Component(object):
+class Component:
 
     def __init__(self, create):
         self.create = create
@@ -357,10 +272,6 @@ class Daemon(_EntryPoint):
         -v, --verbose                           Increase verbosity (DEBUG)
         -q, --quiet                             Decrease verbosity
 
-        -0, --udisks-auto                       Auto discover UDisks version
-        -1, --use-udisks1                       Use UDisks1 as backend
-        -2, --use-udisks2                       Use UDisks2 as backend
-
         -h, --help                              Show this help
         -V, --version                           Show version information
 
@@ -382,19 +293,20 @@ class Daemon(_EntryPoint):
         --password-cache MINUTES                Set password cache timeout
         --no-password-cache                     Disable password cache
 
+        -f PROGRAM, --file-manager PROGRAM      Set program for browsing
+        -F, --no-file-manager                   Disable browsing
+
+        --terminal COMMAND                      Set terminal command line
+                                                (e.g. "termite -d")
+        --no-terminal                           Disable terminal action
+
         -p COMMAND, --password-prompt COMMAND   Command for password retrieval
         -P, --no-password-prompt                Disable unlocking
 
         --notify-command COMMAND                Command to execute on events
         --no-notify-command                     Disable command notifications
-
-    Deprecated options:
-        -f PROGRAM, --file-manager PROGRAM      Set program for browsing
-        -F, --no-file-manager                   Disable browsing
     """
 
-    name = 'udiskie'
-
     option_defaults = extend(_EntryPoint.option_defaults, {
         'automount': True,
         'notify': True,
@@ -402,6 +314,7 @@ class Daemon(_EntryPoint):
         'menu': 'flat',
         'appindicator': False,
         'file_manager': 'xdg-open',
+        'terminal': '',
         'password_prompt': 'builtin:gui',
         'password_cache': False,
         'notify_command': None,
@@ -424,8 +337,6 @@ class Daemon(_EntryPoint):
 
     def _init(self):
 
-        """Implements _EntryPoint._init."""
-
         import udiskie.prompt
 
         config = self.config
@@ -434,18 +345,23 @@ class Daemon(_EntryPoint):
         # prepare mounter object
         prompt = udiskie.prompt.password(options['password_prompt'])
         browser = udiskie.prompt.browser(options['file_manager'])
+        terminal = udiskie.prompt.browser(options['terminal'])
         cache = None
 
-        if options['password_cache'] is not False:
+        try:
             import udiskie.cache
             timeout = int(options['password_cache']) * 60
             cache = udiskie.cache.PasswordCache(timeout)
+        except ImportError:
+            cache = None
 
         self.mounter = udiskie.mount.Mounter(
             config=config.device_config,
             prompt=prompt,
             browser=browser,
+            terminal=terminal,
             cache=cache,
+            cache_hint=options['password_cache'],
             udisks=self.udisks)
 
         # check component availability
@@ -459,7 +375,7 @@ class Daemon(_EntryPoint):
             logging.getLogger(__name__).error(libnotify_not_available)
             options['notify'] = False
 
-        if options['tray'] and not has_Gtk(3) and not _in_X:
+        if options['tray'] and not _in_X:
             no_X_session = _(
                 "Not run within X session. "
                 "\nStarting udiskie without tray icon.\n")
@@ -479,9 +395,9 @@ class Daemon(_EntryPoint):
         # start components
         tasks = []
 
-        self.notify         = Component(self._load_notify)
-        self.statusicon     = Component(self._load_statusicon)
-        self.automounter    = Component(self._load_automounter)
+        self.notify = Component(self._load_notify)
+        self.statusicon = Component(self._load_statusicon)
+        self.automounter = self._load_automounter(options['automount'])
 
         if options['notify']:
             self.notify.activate()
@@ -492,12 +408,11 @@ class Daemon(_EntryPoint):
             self.statusicon.activate()
             tasks.append(self.statusicon.instance._icon.task)
         else:
-            tasks.append(Async())
+            tasks.append(Future())
         if options['automount']:
-            self.automounter.activate()
             tasks.append(self.mounter.add_all())
 
-        return AsyncList(tasks)
+        return gather(*tasks)
 
     def _load_notify(self):
         import udiskie.notify
@@ -522,6 +437,7 @@ class Daemon(_EntryPoint):
     def _load_statusicon(self):
         import udiskie.tray
         options = self.options
+        config = self.config
 
         if options['tray'] == 'auto':
             smart = True
@@ -540,7 +456,8 @@ class Daemon(_EntryPoint):
         else:
             raise ValueError("Invalid menu: %s" % (options['menu'],))
 
-        menu_maker = udiskie.tray.UdiskieMenu(self, icons, actions, flat)
+        menu_maker = udiskie.tray.UdiskieMenu(self, icons, actions, flat,
+                                              config.quickmenu_actions)
         if options['appindicator']:
             import udiskie.appindicator
             TrayIcon = udiskie.appindicator.AppIndicatorIcon
@@ -549,9 +466,9 @@ class Daemon(_EntryPoint):
         trayicon = TrayIcon(menu_maker, icons)
         return udiskie.tray.UdiskieStatusIcon(trayicon, menu_maker, smart)
 
-    def _load_automounter(self):
+    def _load_automounter(self, automount):
         import udiskie.automount
-        return udiskie.automount.AutoMounter(self.mounter)
+        return udiskie.automount.AutoMounter(self.mounter, automount)
 
 
 class Mount(_EntryPoint):
@@ -570,10 +487,6 @@ class Mount(_EntryPoint):
         -v, --verbose                           Increase verbosity (DEBUG)
         -q, --quiet                             Decrease verbosity
 
-        -0, --udisks-auto                       Auto discover UDisks version
-        -1, --use-udisks1                       Use UDisks1 as backend
-        -2, --use-udisks2                       Use UDisks2 as backend
-
         -h, --help                              Show this help
         -V, --version                           Show version information
 
@@ -589,8 +502,6 @@ class Mount(_EntryPoint):
         -P, --no-password-prompt                Disable unlocking
     """
 
-    name = 'udiskie-mount'
-
     option_defaults = extend(_EntryPoint.option_defaults, {
         'recursive': None,
         'options': None,
@@ -607,8 +518,6 @@ class Mount(_EntryPoint):
 
     def _init(self):
 
-        """Implements _EntryPoint._init."""
-
         import udiskie.prompt
 
         config = self.config
@@ -632,7 +541,7 @@ class Mount(_EntryPoint):
                      for path in options['<device>']]
         else:
             tasks = [mounter.add_all(recursive=recursive)]
-        return AsyncList(tasks)
+        return gather(*tasks)
 
 
 class Umount(_EntryPoint):
@@ -651,10 +560,6 @@ class Umount(_EntryPoint):
         -v, --verbose               Increase verbosity (DEBUG)
         -q, --quiet                 Decrease verbosity
 
-        -0, --udisks-auto           Auto discover UDisks version
-        -1, --use-udisks1           Use UDisks1 as backend
-        -2, --use-udisks2           Use UDisks2 as backend
-
         -h, --help                  Show this help
         -V, --version               Show version information
 
@@ -674,8 +579,6 @@ class Umount(_EntryPoint):
         -L, --no-lock               Don't lock device
     """
 
-    name = 'udiskie-umount'
-
     option_defaults = extend(_EntryPoint.option_defaults, {
         'detach': None,
         'eject': False,
@@ -694,8 +597,6 @@ class Umount(_EntryPoint):
 
     def _init(self):
 
-        """Implements _EntryPoint._init."""
-
         config = self.config
         options = self.options
 
@@ -712,7 +613,7 @@ class Umount(_EntryPoint):
                      for path in options['<device>']]
         else:
             tasks = [mounter.remove_all(**strategy)]
-        return AsyncList(tasks)
+        return gather(*tasks)
 
 
 def _parse_filter(spec):
@@ -744,10 +645,6 @@ class Info(_EntryPoint):
         -v, --verbose               Increase verbosity (DEBUG)
         -q, --quiet                 Decrease verbosity
 
-        -0, --udisks-auto           Auto discover UDisks version
-        -1, --use-udisks1           Use UDisks1 as backend
-        -2, --use-udisks2           Use UDisks2 as backend
-
         -h, --help                  Show this help
         -V, --version               Show version information
 
@@ -763,8 +660,6 @@ class Info(_EntryPoint):
                                     filter.
     """
 
-    name = 'udiskie-info'
-
     option_defaults = extend(_EntryPoint.option_defaults, {
         'output': '',
         'filter': '',
@@ -779,8 +674,6 @@ class Info(_EntryPoint):
 
     def _init(self):
 
-        """Implements _EntryPoint._init."""
-
         config = self.config
         options = self.options
 
@@ -801,11 +694,9 @@ class Info(_EntryPoint):
                 return getattr(device, output)
         # new behaviour: format string
         else:
-            from string import Formatter
-            formatter = Formatter()
             def format_output(device):
                 view = ObjDictView(device, DeviceFilter.VALID_PARAMETERS)
-                return formatter.vformat(output, (), view)
+                return output.format_map(view)
 
         filters = [_parse_filter(spec) for spec in options['filter']]
         matcher = DeviceFilter(dict(filters))
@@ -814,4 +705,4 @@ class Info(_EntryPoint):
             if matcher.match(device):
                 print(format_output(device))
 
-        return AsyncList([])
+        return gather()
